/* Ambisonic encoder module for OggEnc
*  Copyright (C) 2007 Sebastian Olter
*
*  This program is free software; you can redistribute it and/or
*  modify it under the terms of the GNU General Public License
*  as published by the Free Software Foundation; either version 2
*  of the License, or (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "pan.h"
#include <stdlib.h>
#include <math.h>
#ifdef WIN32
#include <windows.h>
#endif
#ifndef M_PI
#define M_PI        3.14159265358979323846
#endif
const char *ambichan[] = {
	NULL,
	"W",
	"WY",
	"WXY",
	"WXYZ",
	"WXYRS",
	"WXYZRS",
	"WXYRSPQ",
	"WXYZRSPQ",
	"WXYZRSTUV",
	NULL,
	"WXYZRSTUVPQ",
	NULL,
	NULL,
	NULL,
	NULL,
	"WXYZRSTUVKLMNOPQ"
};

#if defined(_WIN32)
static char* strsep(char** haystack, const char *needle)
{
    char *c, *tmp, *p;
    if (haystack == NULL)
        return NULL;
    for (c = needle; *c != '\0'; c++)
    {
        p = *haystack;
        do
            p++;
        while ((*p != '\0') && (*p != *c));
        tmp = *haystack;
        if (*p != '\0')
        {
            *p = '\0';
            *haystack = p + 1;
        }
        else
            *haystack = NULL;
        return tmp;
    }
    return NULL;
}
#endif

typedef struct
{
    int inch, outch, nbufs;
    float matrix[255][9]; /* inch x outch */
    audio_read_func real_reader;
    void *real_readdata;
    float **bufs;
}
Pan;

static float furse_malham(float A, float E, char channel)
{
    switch (channel)
    {
    case 'W':
        return 1. / sqrt(2);
    case 'X':
        return cos(A) * cos(E);
    case 'Y':
        return sin(A) * cos(E);
    case 'Z':
        return sin(E);
    case 'R':
        return 1.5 * sin(E)*sin(E) - 0.5;
    case 'S':
        return 2./sqrt(3.) * cos(A) * sin(2.*E);
    case 'T':
        return 2./sqrt(3.) * sin(A) * sin(2.*E);
    case 'U':
        return 2./sqrt(3.) * cos(2.*A) * cos(E)*cos(E);
    case 'V':
        return 2./sqrt(3.) * sin(2.*A) * cos(E)*cos(E);
    case 'K':
        return sin(E) * (5. * sin(E)*sin(E) - 3.) / 2.;
    case 'L':
        return sqrt(45./32.) * sqrt(3./8.)*cos(A)*cos(E)*(5.*sin(E)*sin(E)-1.);
    case 'M':
        return sqrt(45./32.) * sqrt(3./8.)*sin(A)*cos(E)*(5.*sin(E)*sin(E)-1.);
    case 'N':
        return 3./sqrt(5.) * sqrt(15./2.) * cos(2.*A) * sin(E) * cos(E)*cos(E);
    case 'O':
        return 3./sqrt(5.) * sqrt(15./2.) * sin(2.*A) * sin(E) * cos(E)*cos(E);
    case 'P':
        return sqrt(8./5.) * sqrt(5./8.) * cos(3.*A) * cos(E)*cos(E)*cos(E);
    case 'Q':
        return sqrt(8./5.) * sqrt(5./8.) * sin(3.*A) * cos(E)*cos(E)*cos(E);
    default:
        return 1.;
    }
}

static int setup_pan_coeffs(oe_enc_opt *opt, const char* channels_in_s)
{
    Pan* pan = opt->readdata;
    int inch = 0;
    float A, E;
    const char *outch;
    char *p, *channels_in = malloc(strlen(channels_in_s) + 1);
    strcpy(channels_in, channels_in_s);
    do
    {
        p = strsep(&channels_in, ",");
        if (strcmp(p, "lfe"))
        {
            A = strtod(p, NULL) * 2. * M_PI / 360.;
            strsep(&p, ":");
            E = (p) ? (strtod(p, NULL) * 2. * M_PI / 360.) : 0.;
            for (outch = ambichan[opt->channels]; outch < ambichan[opt->channels] + opt->channels; outch++)
                pan->matrix[inch][outch - ambichan[opt->channels]] = furse_malham(A, E, *outch);
        }
        else
        {
            A = E = 0.;
            pan->matrix[inch][0] = 1.;
        }
        inch++;
    }
    while (channels_in);
    pan->inch = inch;
    pan->outch = opt->channels;
    return inch;
}

static long read_pan(void *data, float **buffer, int samples)
{
    Pan* pan = data;
    long in_samples = pan->real_reader(pan->real_readdata, pan->bufs, samples);
    int i, x, y;
    float W, Y; // for stereo downmix
    for (i=0; i < in_samples; i++)
    {
        for (y = 0; y < pan->outch; y++)
        {
            buffer[y][i] = 0.;
            for (x = 0; x < pan->inch; x++)
                buffer[y][i] += pan->bufs[x][i] * pan->matrix[x][y] / pan->inch;
        }
        if(pan->outch == 2)
        {
			W = buffer[0][i];
			Y = buffer[1][i];
			buffer[0][i] = W + Y;
			buffer[1][i] = W - Y;
		}
    }
    return in_samples;
}

void setup_pan(oe_enc_opt *opt, const char* channels_in, int channels_out)
{
	char *chan;
    int i, inch = opt->channels;
    Pan *pan = calloc(1, sizeof(Pan));
    if(!ambichan[channels_out])
    {
    	fprintf(stderr,"No suitable Ambisonic configuration for %d channels. Panning disabled.\n",channels_out);
    	return;
	}
	chan = malloc(strlen(ambichan[channels_out]) + 1);
    pan->real_reader = opt->read_samples;
    pan->real_readdata = opt->readdata;
    opt->read_samples = read_pan;
    opt->readdata = pan;


    if (!strcmp(channels_in, "itu5.1"))
        channels_in = "30,330,0,lfe,125,235";
    else if (!strcmp(channels_in, "5.1ac3"))
        channels_in = "30,330,125,235,0,lfe";

    opt->channels = channels_out;
    setup_pan_coeffs(opt, channels_in);

    pan->bufs = malloc(inch * sizeof(float*));
    for (i = 0; i < inch; i++)
        pan->bufs[i] = malloc(4096 * sizeof(float));
	pan->nbufs = inch;

    pan->inch = min(pan->inch, inch); /* Avoid reading from non-existing buffers */
    fprintf(stderr, "Enabling Ambisonic encoder %dch->", pan->inch);
    switch(channels_out)
    {
    	case 1:
			fprintf(stderr, "mono\n");
			break;
    	case 2:
			fprintf(stderr, "stereo\n");
			break;
    	default:
			fprintf(stderr, "%s\n", ambichan[channels_out]);
			break;
	}
	free(chan);
}

void clear_pan(oe_enc_opt *opt)
{
    int i;
    Pan* pan = opt->readdata;
    opt->read_samples = pan->real_reader;
    opt->readdata = pan->real_readdata;
    for (i = 0; i < pan->nbufs; i++)
        free(pan->bufs[i]);
    free(pan->bufs);
    free(pan);
}